/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/main.c
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 * This is the main file for the gtk+ client.
 *
 * GUI TODO: 
 * 
 * - Most of the GUI code is awful write-once-read-none-spaghetti
 *   => should be somehow restructured and/or cleaned up.
 * - Anticipating future off-band sharing of gnunet keys,
 *   add possibility to download files by specifying their keys
 *   manually (by copy-pasting - a URI for gnunet keys should be
 *   standardized!).
 **/

#include "config.h"
#include <gtk/gtk.h>
#include <glib.h>
#include <unistd.h>
#include <sys/types.h>

#include "util/getopt.h"
#include "util/initialize.h"
#include "util/storage.h"
#include "util/logging.h"
#include "util/initialize.h"
#include "applications/afs/afs.h"
#include "applications/afs/gtkui/helper.h"
#include "applications/afs/gtkui/main.h"
#include "applications/afs/gtkui/search.h"
#include "applications/afs/gtkui/insert.h"
#include "applications/afs/encoding/requestmanager.h"
#include "applications/afs/encoding/searchutil.h"
#include "applications/afs/encoding/insertutil.h"
#include "applications/afs/util/policy.h"
#include "applications/common/printhelp.h"
#include "util/configuration.h"

/* a dastardly unsafe cheese kludge */
int _I_AM_DEAD=NO;

static GtkWidget * main_window_input_line = NULL;
static GtkNotebook * notebook = NULL;

#define ABOUT_STRING "\nGNUnet "\
  VERSION\
  " gnunet-gtk "\
  AFS_VERSION\
  "\n\n\n"\
  "This is free software, released under GNU General Public License version 2."\
  "\n\n\n"\
  "For more information, visit the GNUnet homepage at \n\n"\
  "http://www.ovmj.org/GNUnet/\n"

/**
 * Remove the active page from the search results notebook.
 * The respective search will be stopped as well
 * (by a callback assigned to the page earlier on).
 **/
static void close_search(GtkButton * button,
			 gpointer dummy) {
  gint page;
  
  page = gtk_notebook_get_current_page(notebook);
  gtk_notebook_remove_page(notebook, page);
  /* Need to refresh the widget --
     This forces the widget to redraw itself. */
  gtk_widget_draw(GTK_WIDGET(notebook), NULL);
}

/**
 * This displays an about window
 **/
static void about(GtkWidget *dummy,
                  gpointer data) {
  GtkWidget *window, *box1, *table, *text, *button;

  /* create about window */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect(GTK_OBJECT(window), 
                     "delete_event",
                     GTK_SIGNAL_FUNC(deleteEvent), 
		     NULL);
 
  gtk_window_set_title(GTK_WINDOW(window), "About gnunet-gtk");
  gtk_widget_set_usize(GTK_WIDGET(window), 600, 300);

  box1 = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER (window), box1);
  gtk_widget_show(box1);

  table = gtk_table_new(2, 2, FALSE);
  gtk_table_set_row_spacing(GTK_TABLE (table), 0, 2);
  gtk_table_set_col_spacing(GTK_TABLE (table), 0, 2);
  gtk_box_pack_start(GTK_BOX (box1), table, TRUE, TRUE, 0);
  gtk_widget_show(table);

  /* create a text widget */
  text = gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT (text), FALSE);
  gtk_table_attach(GTK_TABLE (table), text, 0, 1, 0, 1,
		   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  gtk_widget_show(text);
  gtk_widget_realize(text);

  /* write some about text */
  gtk_text_freeze(GTK_TEXT (text));

  gtk_text_insert(GTK_TEXT(text), 
		  NULL, 
		  &text->style->black, 
		  NULL,
		  ABOUT_STRING, -1); 
  
  gtk_text_thaw(GTK_TEXT(text));

  /* finish with a close button */
  button = gtk_button_new_with_label("Right");
  gtk_box_pack_start(GTK_BOX (box1), button, FALSE, FALSE, 0);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(destroyWidget), window);
  gtk_widget_show(button);
  gtk_widget_show(window);
}

/**
 * This method is called whenever the user clicks the
 * search button of the main window. 
 * 
 * @param notebook the notebook where the opened result window will be put.
 **/
static void search(GtkWidget * widget,
		   GtkNotebook * notebook) {
  gchar * searchString;
  GtkWidget * label, * tmp;
  int len,i; 

  searchString 
    = gtk_entry_get_text(GTK_ENTRY(main_window_input_line));

  if(searchString == NULL) {
    guiMessage("searchString == NULL\n");
    return;
  }

  // Remove & heading spaces (parsing bugs if not)
  len = strlen(searchString);
  for(i = len-1;(searchString[i]==' ') && (i>=0);i--) 
    searchString[i]=0;
  while(*searchString==' ')
    searchString++;

  if(searchString[0]==0) {
    guiMessage("No search key given!\n");
    return;
  }

  tmp = getSearchWindow(searchString);
  if(tmp != NULL) {
    label = gtk_label_new(searchString);
    gtk_notebook_append_page(notebook, tmp, label);
    gtk_widget_show(tmp);
  }
  gtk_entry_set_text(GTK_ENTRY(main_window_input_line), 
		     ""); 
}

/**
 * Exit the application 
 **/
static void destroy(GtkWidget * widget,
		    gpointer data) {
  _I_AM_DEAD=YES;
  if(notebook != NULL) {
    int i=0;
    
    while(gtk_notebook_get_nth_page(notebook, 0) != NULL) {
      LOG(LOG_DEBUG, "DEBUG: Removing search page %d\n", i++);
      gtk_notebook_remove_page(notebook, 0);
    }
  }

  gtk_main_quit(); 
  /* exit(0); */
}

static void destroy_stub(void) {
  destroy(NULL, NULL);
}

/**
 * The pulldown menus
 **/
static GtkItemFactoryEntry menu_items[] = {
  { "/_File",         NULL,         NULL, 0, "<Branch>" },
  { "/File/_Insert",   "<control>I", openSelectFile, 0, NULL },
  { "/File/sep1",     NULL,         NULL, 0, "<Separator>" },
  { "/File/_Abort search",         "<control>A", close_search, 0, NULL },
  { "/File/sep1",     NULL,         NULL, 0, "<Separator>" },
  { "/File/_Quit",     "<control>Q", destroy_stub, 0, NULL },
/*
  { "/_Options",      NULL,         NULL, 0, "<Branch>" },
  { "/Options/Preferences(not impl)",  NULL,         NULL, 0, NULL },
*/
  { "/_Help",         NULL,         NULL, 0, "<LastBranch>" },
  { "/_Help/About",   NULL,         about, 0, NULL },
};

/**
 * This creates the main window
 **/
static void makeMainWindow() {
  GtkWidget * window;
  GtkWidget * button;
  GtkWidget * hbox, * vbox;
  GtkWidget * menubar;
  GtkWidget * table;
  GtkWidget * label;
  GtkAccelGroup * accel_group;
  GtkItemFactory * item_factory;
  gint nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
 
  gtk_window_set_title(GTK_WINDOW(window), "GNUnet: gtk+ GUI");
  gtk_widget_set_usize(GTK_WIDGET(window), 800, 300);
  vbox = gtk_vbox_new(FALSE, 1);
  gtk_container_add(GTK_CONTAINER(window), vbox);
  gtk_signal_connect(GTK_OBJECT(window), 
		      "delete_event",
		      GTK_SIGNAL_FUNC(deleteEvent), NULL);
  gtk_signal_connect(GTK_OBJECT(window), 
		      "destroy",
		      GTK_SIGNAL_FUNC(destroy), NULL);
  gtk_widget_show(vbox);
  
  /* create pulldown menues */
  accel_group = gtk_accel_group_new ();
  item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>",
                       accel_group);
  gtk_item_factory_create_items(item_factory, nmenu_items, menu_items, NULL);
  gtk_window_add_accel_group(GTK_WINDOW (window), accel_group);
  menubar = gtk_item_factory_get_widget(item_factory, "<main>");
  gtk_box_pack_start(GTK_BOX (vbox), menubar, FALSE, TRUE, 0);
  gtk_widget_show(menubar);

  /* a table to put the search results notebook to */
  table = gtk_table_new(6, 6, TRUE);
  gtk_box_pack_start(GTK_BOX (vbox), table, TRUE, TRUE, 0);
  gtk_widget_show(table);
  
  /* add "notebook" for search results */
  notebook = GTK_NOTEBOOK(gtk_notebook_new());
  gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
  gtk_table_attach_defaults(GTK_TABLE(table), GTK_WIDGET(notebook), 0,6,0,6);
  gtk_widget_show(GTK_WIDGET(notebook));

  /* At the bottom, put the search box */
  hbox = gtk_hbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  gtk_widget_show(hbox);
 
  /* search entry label */
  label = gtk_label_new("Keyword(s):");
  gtk_box_pack_start(GTK_BOX(hbox),
  			label, FALSE, FALSE, 0);
  gtk_widget_show(label);

  /* search entry */
  main_window_input_line = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(hbox), 
		     main_window_input_line, TRUE, TRUE, 0);
  gtk_signal_connect(GTK_OBJECT(main_window_input_line),
		     "activate",
                     GTK_SIGNAL_FUNC(search),
                     notebook);
  gtk_widget_show(main_window_input_line);

  /* search button */
  button = gtk_button_new_with_label("Search");
  gtk_signal_connect(GTK_OBJECT (button), 
		     "clicked",
		     GTK_SIGNAL_FUNC(search), notebook);
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  /* show the main window */
  gtk_widget_show(window);
}

/**
 * Perform option parsing from the command line. 
 **/
static int parseOptions(int argc, 
			char ** argv) {
  int shouldexit = 0;
  int c;

  while (1) {
    int option_index=0;
    static struct option long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { 0,0,0,0 }
    };
    
    c = getopt_long(argc,
		    argv, 
		    "vhdc:L:H:", 
		    long_options, 
		    &option_index);
    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, optarg))
      continue;    
    switch(c) {
    case 'v': 
      printf("GNUnet v%s, gnunet-gtk v%s\n",
	     VERSION,
	     AFS_VERSION);
      shouldexit = 1;
      break;
    case 'h': 
      printf("USAGE: gnunet-gtk [OPTIONS]\n");
      printHelpDefault();
      shouldexit = 1;
      break;
    default:
      LOG(LOG_FAILURE,
	  "Unknown option %c. Aborting.\n"
	  "Use --help to get a list of options.\n",
	  c);
      shouldexit=1;    
    } /* end of parsing commandline */
  }
  if (optind < argc) {
    LOG(LOG_WARNING, "Invalid arguments: ");
    while (optind < argc)
      LOG(LOG_WARNING, "%s ", argv[optind++]);
    errexit("Bad arguments.");
  }
  return shouldexit;
}

/**
 * The main function.
 **/
int main(int argc,
         char * argv[]) {
  char * downloadDir;

  if (SYSERR == initUtil(argc, argv, &parseOptions))
    return(0);
  startCron();
  initAnonymityPolicy(NULL);

  g_thread_init(NULL);
  /*  gdk_threads_init(); */
  gtk_init(&argc, &argv);

  /* Create and enter download dir */
  downloadDir = getConfigurationString("GNUNET-GTK",
  				       "DOWNLOADDIR");
  if(downloadDir != NULL) {
    mkdirp(downloadDir);
    chdir(downloadDir);
  }
  xfree(downloadDir,
	"freeing download directory");
  makeMainWindow();
  gtk_main();

  LOG(LOG_DEBUG, 
      "DEBUG: GUI leaving...\n");

  doneAnonymityPolicy();
  stopCron();
  doneUtil();
  return 0;
}

/* end of main.c */
